//
// Copyright (c) 2012-2016 Michele Segata <segata@ccs-labs.org>, Tobias Hardes <hardes@ccs-labs.org>
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program.  If not, see http://www.gnu.org/licenses/.
//

#include "veins/modules/application/platooning/protocols/HumanInterferingProtocol.h"

#include "veins/modules/application/platooning/messages/LaneChangeBasicSafetyMessage_m.h"
#include "veins/modules/application/platooning/messages/InterferingBeacon_m.h"
#include "veins/modules/application/platooning/messages/ManeuverMessage_m.h"
#include "veins/modules/mobility/traci/TraCICoord.h"



#include "veins/modules/messages/PhyControlMessage_m.h"

Define_Module(HumanInterferingProtocol)

void HumanInterferingProtocol::initialize(int stage) {

	BaseApplLayer::initialize(stage);

	if (stage == 0) {
	    id = getParentModule()->getId();
	    WATCH(id);
	    WATCH(currentLane);
		//init class variables
		sendBeacon = 0;

		//get gates
		lowerLayerIn = findGate("lowerLayerIn");
		lowerLayerOut = findGate("lowerLayerOut");

		//get traci interface
		mobility = Veins::TraCIMobilityAccess().get(getParentModule());
		traci = mobility->getCommandInterface();
		traciVehicle = mobility->getVehicleCommandInterface();

		//get pointer to mac
		mac = FindModule<Mac1609_4 *>::findSubModule(getParentModule());

		//tell the unicast protocol below which mac address to use via control message
		UnicastProtocolControlMessage *setMacAddress = new UnicastProtocolControlMessage("");
		setMacAddress->setControlCommand(SET_MAC_ADDRESS);
		//set a mac address not interfering with platooning vehicles
		setMacAddress->setCommandValue(getParentModule()->getIndex() + 1e6);
		send(setMacAddress, lowerControlOut);

		//beaconing interval in seconds
		beaconingInterval = SimTime(par("beaconingInterval").doubleValue());

		//platooning message packet size
		packetSize = par("packetSize").longValue();
		//priority of platooning message
		priority = par("priority").longValue();
		ASSERT2(priority >= 0 && priority <= 7, "priority value must be between 0 and 7");
		//tx power
		txPower = par("txPower").doubleValue();
		//bit rate
		bitrate = par("bitrate").intValue();

		//init messages for scheduleAt
		sendBeacon = new cMessage("sendBeacon");
		scheduleAt(simTime() + uniform(0, beaconingInterval), sendBeacon);
	}

	if (stage == 1) {
		//METHOD 1: setting tx power and bitrate for all frames sent
		//call this method at stage 1 otherwise the MAC might overwrite the
		//values with the ones loaded from omnetpp.ini
		mac->setTxPower(txPower);
		mac->setMCS(getMCS(bitrate, BW_OFDM_10_MHZ));
		double releaseTime = 1.2;
		if(traciVehicle->getLength()>4){
            traciVehicle->changeLane(1, 0);
            traciVehicle->setLaneChangeMode(512);
        }
        else {
            if (simTime().dbl() < releaseTime) {
                if (traciVehicle->getLaneIndex() == 1) {
                    traciVehicle->changeLane(0, 2);
                    traciVehicle->setFixedLane(traciVehicle->getLaneIndex(), false);
                }
                scheduleAt(releaseTime, releaseFixedLane);
                laneChanged = false;
            }
        }


	}

}

HumanInterferingProtocol::~HumanInterferingProtocol() {
	cancelAndDelete(sendBeacon);
	sendBeacon = nullptr;
}


void HumanInterferingProtocol::handleMsg(cMessage *msg) {
    delete msg;
}
void HumanInterferingProtocol::handleSelfMsg(cMessage *msg) {
    if (msg == sendBeacon) {
        if(traciVehicle->getLength()<5){
            if (traciVehicle->getLaneIndex() == 1 && simTime().dbl() < 5) {
                int lane = -1;
                do {
                    lane = intuniform(0, numLanes);
                } while (lane == 1); //lane 1 is the platoon/truck lane, wee keep this free until the platoon appears
                traciVehicle->changeLane(lane, 0);
                traciVehicle->setFixedLane(lane, false);
            }
        }
        sendBSM();
        scheduleAt(simTime() + beaconingInterval, sendBeacon);
    }
    if(msg == releaseFixedLane) {
        traciVehicle->setFixedLane(-1, false); // Lane changeing gets random again
    }

}

void HumanInterferingProtocol::sendBSM() {


    //create and send beacon
    UnicastMessage *unicast = new UnicastMessage("", INTERFERENCE_TYPE);
    unicast->setDestination(-1);
    unicast->setPriority(priority);
    unicast->setChannel(Channels::CCH);
    unicast->setKind(1337); // ID for the protocol

    LaneChangeBasicSafetyMessage* bsm = new LaneChangeBasicSafetyMessage();
    bsm->setPositionX(mobility->getCurrentPosition().x);
    bsm->setPositionY(mobility->getCurrentPosition().y);

    bsm->setTimestamp(simTime());
    bsm->setLane(traciVehicle->getLaneIndex());
    bsm->setVehicleId(id);
    bsm->setVehicleSize(traciVehicle->getLength());
    bsm->setSpeed(mobility->getCurrentSpeed().x);

    //put platooning beacon into the message for the UnicastProtocol
    unicast->encapsulate(bsm);
    sendDown(unicast);
}

void HumanInterferingProtocol::handleLowerMsg(cMessage *msg) {
    if(UnicastMessage* uc = dynamic_cast<UnicastMessage*>(msg)) {
        cMessage* inUc = uc->decapsulate();

        if(traciVehicle->getLength()<5){
            // Fixing the lane until the platoon changed the lane
            if (ManeuverMessage* maneuver = dynamic_cast<ManeuverMessage*>(inUc)) {
                traciVehicle->setFixedLane(traciVehicle->getLaneIndex(), false);
                if (!releaseFixedLane->isScheduled()) {
                    scheduleAt(simTime() + 2, releaseFixedLane);
                }
            }
        }
        delete inUc;
        delete msg;
    }
}
